
/*
   main.c -- the top level caller of pmail
   The TABLE STOP is set to 4 characters in this file.
	
   Copyright (C) 1998, 1999  Xuming <xuming@bigfoot.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include	<stdio.h>
#include	<string.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	<getopt.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<fcntl.h>
#include	<signal.h>
#include	"defs.h"
#include	"pmalloc.h"
#include	"pop3.h"
#include	"smtp.h"


#define		WORKLIST	10


struct		PARA_LIST	{
	char	*optName;
	void	(*action)(char *p);
};


/* these varieties are set by command line arguments */

static	char	*_altHost		= NULL;
static	char	*_altMailbox 	= NULL;
static	int		*_downloadSet 	= NULL;
static	int 	_defAction 		= POP3_NONE;


/* varieties set according to RC file */

static	int		_timeout = 0;
static	int		_verbose = 0;


/* pmail might do a serial of work according to this working list */

static	int 	_schedule [ WORKLIST ] = { 0 };


static void Option(int argc, char **argv);
static void parseOptionFile();
static void setSchedule(int work);
static void addSchedule(int work);
static void setAltHost(char *s);
static void setAltMailbox(char *s);
static void setDnldSet(char *s);
static void setTimeout(char *s);
static void setVerbose(char *s);
static void setDefault(char *s);
static void signalQuit(int sig);
static void appendTemplate();
static void usage(int flag);

#ifdef	DEBUG
static void mainVarietyExposure();
#endif


static	struct	PARA_LIST	ParaList[] = {
	
	{ "default",		setDefault },
	
	{ "debuglevel", 	dbgLevelIs },
	{ "debugdevice",	dbgMsgSendInto },
	{ "logfile",		logFileNameIs },
	{ "loglimit",		logFileLimitedIn },
	
	{ "pop3host",		popSetHost },
	{ "topnumber",		popListTopLines },
	{ "topstyle",		popTopMethod },
	{ "sizelimit",		popSizeLimitation },
	{ "logindelay",		popLoginDelay },
	
	{ "verbose",		setVerbose },
	{ "timeout",		setTimeout },
	{ "default",		setDefault },

	{ "smtp",			smtpSetHost },
	{ "outbox",			smtpSetOutbox },
	{ "bakbox",			smtpSetBackupbox },
	{ "pipeto",			smtpSetPipeto },
	{ "sender",			smtpSetSender },
	{ "handlesent",		smtpSetDowithSent },
	{ "signature",		smtpSetSignature },
	{ "header",			smtpSetExtraHeader },
	
	{ NULL, NULL }
};


int main(int argc, char **argv)
{
	int 	i, rs;

	setAllocHandle(ReleaseResource);
	signal(SIGHUP, signalQuit);
	signal(SIGINT, signalQuit);
	signal(SIGQUIT, signalQuit);
	signal(SIGTERM, signalQuit);
	signal(SIGPIPE, signalQuit);
	signal(SIGALRM, signalQuit);
	
	umask(077);
	Option(argc, argv);
	parseOptionFile();
/*	setlinebuf(stdout);  */
	dbgInit();
	
#ifdef		DEBUG
	
	popVarietyExposure();
	dbgVarietyExposure();
	mainVarietyExposure();
	smtpVarietyExposure();
	
#endif
	
	for (i = 0; i < _schedule[0]; i++)  {
		switch (_schedule[i+1])  {
			case GETMAIL:
				pop3main();
				break;
				
			case SENDMAIL:
				rs = smtpMain(SENDMAIL);
				if (rs != OK)  errReport(rs, "smtp");
				break;
							
			case TEMPLATE:
				appendTemplate();
				break;
				
			case STREAM:
				rs = smtpMain(STREAM);
				if (rs != OK)  errReport(rs, "smtp");
				break;
				
			case HELP:
				usage(HELP);
				break;
				
			case VERSION:
				usage(VERSION);
				break;
		}
	}
	
	ReleaseResource();
	return 0;
}


static void Option(int argc, char **argv)
{
	int 	oidx, c;
	char	*sopt="votf:h:d:s:c:b:l:";
	struct	option	lopt[] = {
		{ "verbose", 0, NULL, 'v' },
		{ "host", 1, NULL, 'h' },
		{ "do", 1, NULL, 'd' },
		{ "list", 1, NULL, 'l' },
		{ "template", 0, NULL, 't' },
		{ "help", 0, NULL, HELP },
		{ "version", 0, NULL, VERSION },
		{ 0, 0, 0, 0 }
	};

	while (1)  {
		c = getopt_long(argc, argv, sopt, lopt, &oidx);
		if (c == -1)  break;
		switch (c)  {
			case HELP:
				setSchedule(HELP);
				break;
			
			case VERSION:
				setSchedule(VERSION);
				break;
			
			case 'v':
				_verbose = YES;
				break;
			
			case 't':
				setSchedule(TEMPLATE);
				break;
			
			case 'o':
				setSchedule(SENDMAIL);
				smtpEnforced();		/* see the routine in smtp.c */
				break;
			
			case 'h':
				setAltHost(optarg);
				break;
			
			case 'l':
				setDnldSet(optarg);
				break;
			
			case 'f':
				setAltMailbox(optarg);
				break;
			
			case 's':
				smtpSetSubject(optarg);
				break;
			
			case 'c':
				smtpSetCc(optarg);
				break;
			
			case 'b':
				smtpSetBcc(optarg);
				break;
						
			case 'd':
				switch (*optarg)  {
					case 'c':
						_defAction = POP3_CK;
						break;
			
					case 't':
						_defAction = POP3_TOP;
						break;
				  
					case 'd':
						_defAction = POP3_DL;
						break;
				  
					case 'f':
						_defAction = POP3_FETC;
						break;
				
					case 'r':
						_defAction = POP3_RM;
						break;
				  
					default:  _defAction = POP3_NONE;
				}
				setSchedule(GETMAIL);
				break;
			
			default:
				setSchedule(HELP);
		}
	}	
	if (_schedule[0])  return;
	if (optind < argc)  {
		smtpSetToaddr(argc - optind, &argv[optind]);
		setSchedule(STREAM);
	}
}


static void parseOptionFile()
{
	FILE    *fp;
	char    buf[SVRBUF], *p;
	struct	PARA_LIST	*pl;
			 
	getAbsolutePath(RCFILE, buf);
	if ((fp = fopen(buf, "rt")) == NULL)  {
		conPrintf("pmail: cannot find option file [%s]\n", buf);
		conPrintf("pmail: check your pmail distribution ");
		conPrintf("package for a sample file\n");
		return;
	}
	
	while (!feof(fp)) {
		
		if (fgets(buf, SVRBUF, fp) == NULL)  break;
		if (availOptionLine(buf) == 0)  continue;
		
		for (pl = ParaList; pl->optName; pl++)  {
			
			if (!strncmp(buf, pl->optName, strlen(pl->optName)))  {
				p = buf + strlen(pl->optName);
				p = SkipLeadSps(p);
				pl->action(p);
				pl = NULL;
				break;
			}
		}
		if (pl)  
			conPrintf("pmail: Warning: unknown option [%s]\n", buf);
	}
}
	


/************************************************************/

static void setSchedule(int work)
{
	_schedule[0] = 1;
	_schedule[1] = work;
}

static void addSchedule(int work)
{
	if (_schedule[0] < (WORKLIST - 1))  {
		_schedule[0]++;
		_schedule[*_schedule] = work;
	}
}

/********************************************************/

static void setAltHost(char *s)
{
	pfree(_altHost);
	_altHost = pmalloc(strlen(s)+1, NULL);
	strcpy(_altHost, s);
}

static void setAltMailbox(char *s)
{
	pfree(_altMailbox);
	_altMailbox = pmalloc(strlen(s)+1, NULL);
	strcpy(_altMailbox, s);
}



/* store and adjust the mail download/handle set. e.g.
   sour:          1,2,4-9,10,12         -->
   download set:  1,2,4,0,9,10,12,-1    (int array)
*/
static void setDnldSet(char *s)
{
	char	buf[SVRBUF], *p;
	int 	*dl;
	
	pfree(_downloadSet);
	_downloadSet = pmalloc(strlen(s) * sizeof(int) * 2, NULL);
	dl = _downloadSet;
	strcpy(buf, s);
	
	for (p = strtok(buf, ","); p != NULL; p = strtok(NULL, ","))  {
		*dl++ = atoi(p);
		s = strchr(p, '-');
		if (s)  *dl++ = 0, *dl++ = atoi(s+1);
	}
	*dl = EOF;		/* -1, end flag */
}


static void setTimeout(char *s)
{
	_timeout = atoi(s);
}

static void setVerbose(char *s)
{
	_verbose = YES;
}

static void setDefault(char *s)
{
	char	*p;
	
	/* this function read a option, 'default xxx,xxx,...', from 
	   RC file. If there were another action specified by command
	   line arguments, the 'default' option should be ignored. */
	
	if (_schedule[0])  return;
	
	if ((p = strtok(s, ",")) == NULL)  return;
	do {
		if (!strcmp(p, "pop3"))			addSchedule(GETMAIL);
		else if (!strcmp(p, "sbox"))	addSchedule(SENDMAIL);
		else if (!strcmp(p, "templ"))	addSchedule(TEMPLATE);
		else if (!strcmp(p, "version")) addSchedule(VERSION);
		else if (!strcmp(p, "help"))	addSchedule(HELP);
	} while ((p = strtok(NULL, ",")) != NULL);
}

static void signalQuit(int sig)
{
	ReleaseResource();
	switch(sig)  {
		case SIGHUP:
			printf("\nEncountered signal SIGHUP (%d)\n", sig);
			break;
			
		case SIGINT:
			printf("\nEncountered signal SIGINT (%d)\n", sig);
			break;
			
		case SIGQUIT:
			printf("\nEncountered signal SIGQUIT (%d)\n", sig);
			break;
			
		case SIGTERM:
			printf("\nEncountered signal SIGTERM (%d)\n", sig);
			break;
			
		case SIGALRM:
			printf("\nEncountered timeout signal (%d)\n", sig);
			break;
			
		case SIGPIPE:
			printf("\nEncountered signal SIGPIPE (%d)\n", sig);
			break;
	}
	exit(sig);
}


static void appendTemplate()
{
	FILE	*fp;
	char	*outbox, *sender;
	
	outbox = _altMailbox;
	if (outbox == NULL)  outbox = smtpGetOutbox();
	
	fp = fopen(outbox, "at");
	if (fp == NULL)  fp = stdout;
	
	if ((sender = smtpGetSender()) == NULL)  {
		fprintf(fp, "\nFrom - %s\n", timenow());
		fprintf(fp, "From: \n");
	} else {
		fprintf(fp, "\nFrom %s %s\n", emailAddr(sender), timenow());
		fprintf(fp, "From: %s\n", sender);
	}
	fprintf(fp, "To: \n");
	fprintf(fp, "Subject: \n\n");
	if (fp != stdout)  fclose(fp);
	printf("Simple template is appended into %s.\n",
			outbox ? outbox : "stdout");
}
	


#ifdef		DEBUG

static void mainVarietyExposure()
{
	int		i;
	
	dbgPrintf("Alternative host is [%s]\n", _altHost);
	dbgPrintf("Alternative mailbox is [%s]\n", _altMailbox);
	dbgPrintf("Default action is [%d]\n", _defAction);
	dbgPrintf("Time out is set to [%d]\n", _timeout);
	dbgPrintf("Verbose switch is [%d]\n", _verbose);
	
	dbgPrintf("Download mail set is: ");
	if (_downloadSet)  for (i=0; _downloadSet[i]>=0; i++)  {
		dbgPrintf2(1, "%d ", _downloadSet[i]);
	} else  dbgPrintf2(1, "NONE");
	dbgPrintf2(1, "\n");
	
	dbgPrintf("Job schedule is: ");
	for (i=0; i<_schedule[0]; i++)  dbgPrintf2(1, "%d ", _schedule[i]);
	dbgPrintf2(1, "\n");
}

#endif
	
/**************************************************************/
		
static void usage(int flag)
{
	char *ver="pmail %s, by Xuming <xuming@bigfoot.com>\n";

	char *msg = "\
Usage:\n\
 pmail [--help] [--version] [-v] [-f filebox] [-h host] [-l ...]\n\
\n\
 pmail -dd, --do d  download mails and remove them from server\n\
 pmail -df, --do f  fetch mails and reserve them in the server\n\
 pmail -dc, --do c  check mailbox status in the server\n\
 pmail -dt, --do t  print header or other informations only\n\
 pmail -dr, --do r  remove mails from server\n\
\n\
 pmail -o   send mails stored in the outgoing mailbox\n\
\n\
 pmail [-s subject] [-c cc-addr] [-b bcc-addr] to-addr [...]\n\
\n\
 pmail -t, --template    appends a template to the outgoing mailbox\n\
";
   
	if (flag == VERSION)  printf(ver, PMAILVER); else  printf(msg);
}
	
/**********************************************************/

char *getAltHost()
{
	return _altHost;
}

char *getAltMailbox()
{
	return _altMailbox;
}

int isVerbose()
{
	return _verbose;
}

int defaultPop3Action()
{
	return _defAction;
}
	
int timeout()
{
	return _timeout;
}

int *isSetExist()
{
	return _downloadSet;
}


/* NOTES that TRUE will be return when NULL set encountered !! */

int isInSet(int obj)
{
	int 	i, *set;
	
	set = _downloadSet;
	if (set == NULL)  return YES;
	for (i = 0; set[i] != EOF; i++)  {
		if (set[i] == 0)  {
			if ((i == 0) || (set[i+1] == EOF))  continue;
			else {
				if ((obj > set[i-1]) && (obj < set[i+1]))  return YES;
				if ((obj < set[i-1]) && (obj > set[i+1]))  return YES;
			}
		} else if (obj == set[i])  return YES;
	}
	return NO;
}
		

/* this procedure used to release all of the source which pmail ever
   occured. It also is looked on the last defend line as signal 
   encountered or other urgent affair occured */

void ReleaseResource()
{
	popReleaseAll();
	dbgClose();
	smtpRelease();
	pfree(_altHost);
	pfree(_altMailbox);
	pfree(_downloadSet);
}

