/*
   misc.c -- miscellany functions
   The TABLE STOP is set to 4 characters in this file.
	
   Copyright (C) 1998, 1999  Xuming <xuming@bigfoot.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include	<stdio.h>
#include	<string.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	<time.h>
#include	"defs.h"


int isblankline(char *buf)
{
	while (*buf)  if (*buf > ' ')  return 0;  else  buf++;
	return 1;
}


char *SkipLeadSps(char *p)
{
	for ( ; *p && (*p <= ' '); p++);
	return p;
}


char *EatSpsFromTail(char *p)
{
	if (*p)  {
		p = p + strlen(p) - 1;
		for (; *p <= ' '; *p-- = '\0');
	}
	return p;
}
	
	
/* this function is used to handling the RC file.
   it ought to remove the commentary part, the leading and ending
   space in the specified string. e.g. ('<>' means SPACE or less)
   
       <><><>smtp smtp.host<><><># specify a SMTP server\n   --->
	   smtp smtp.host
*/

int availOptionLine(char *buf)
{
	char	tmp[SVRBUF], *availBegin, *p;
	
	availBegin = SkipLeadSps(buf);
	
	if (*availBegin == COMMENTDELM)  {
		*availBegin = '\0';
	} else if (*availBegin)  {
		p = availBegin + 1;
		while ((p = strchr(p, COMMENTDELM)) != NULL)  {
			if (*(p-1) != '\\')  break;
		}
		if (p && (*p == COMMENTDELM))  *p = '\0';
		EatSpsFromTail(availBegin);
	}
	strcpy(tmp, availBegin);
	strcpy(buf, tmp);
	return strlen(buf);
}
	

int strnicmp(char *sour, char *dest, int leng)
{
	int     i;
	char    a, b;
			
#ifdef  strncasecmp
			
	return strncasecmp(sour, dest, leng);
				
#else
				
	for (i=0; i<leng; i++, sour++, dest++)  {
		if ((*sour < 'a') || (*sour > 'z'))  a = *sour;
		else  a = *sour - 'a' + 'A';
		if ((*dest < 'a') || (*dest > 'z'))  b = *dest;
		else  b = *dest - 'a' + 'A';
		if (a != b)  return 1;
		if (a == '\0')  break;
	}
	return 0;
						
#endif
}



/* 
   if there is envirnment variety in the specified path, we should
   change it into absolute path. Note that '~' means home directory.
   i.e,  if home directory is /home/myhome:
   
      sour:                      dest:
   
      ~/Mail               ---> /home/myhome/Mail
	  $HOME/Mail/.pmailrc  ---> /home/myhome/Mail/.pmailrc
*/	  

void getAbsolutePath(char *sour, char *dest)
{
	char    buf[SVRBUF], *p, *env;
	int		rule;

	strcpy(buf, sour);
	if (*buf == '/')  strcpy(dest, "/");  
	else  *dest = '\0';
	
	if ((p = strtok(buf, "/")) == NULL)  return;
	do {
		if (isblankline(p))  continue;
		
		if (!strcmp(p, "~"))  {
			env = getenv("HOME");
			if (env)  strcat(dest, env);
		} else if (*p == '$')  {
			env = getenv(++p);
			if (env)  strcat(dest, env);
		} else  strcat(dest, p);
		
		strcat(dest, "/");
		
	} while ((p = strtok(NULL, "/")) != NULL);
	
	rule = strlen(dest) - 1;
	if (dest[rule] == '/')	dest[rule] = '\0';
}


/* get email address from verbose address. i.e
   
   "XUming" <xuming@bigfoot.com>   --> xuming@bigfoot.com
*/
   
char *emailAddr(char *faddr)
{
	char    *sour, *dest;
	static  char    buf[SVRBUF];
	int 	qflag;
			
	sour = faddr;
	dest = buf;
	qflag = 0;
	while (*sour)  {     /* rfc822, the commentary in the address */
		if (*sour == '(')  qflag++;
		if (*sour == ')')  qflag--;
		
		if (!qflag && (*sour != ')'))  *dest++ = *sour++;
		else sour++;
	}
	*dest = '\0';
	
	if ((sour = strchr(buf, '<')) == NULL)  return buf;
	dest = ++sour;
	if ((sour = strchr(dest, '>')) == NULL)  return dest;
	*sour = '\0';
	return dest;
}


/* get current time like "Tue Jan 26 02:35:07 1999" */

char *timenow()
{
	time_t  tm;
	static  char    buf[32];

	time(&tm);
	strcpy(buf, ctime(&tm));
	chop(buf);
	return buf;
}

/* print strings to control without buffer */

void conPrintf(char *fmt, ...)
{
	char	buf[SVRBUF];
	void	*valist;
	
	valist = &fmt + 1;
	vsnprintf(buf, SVRBUF, fmt, valist);
	write(1, buf, strlen(buf));
}


