
/*
   pmalloc.c -- some function about allocate memory
   The TABLE STOP is set to 4 characters in this file.
	
   Copyright (C) 1998, 1999  Xuming <xuming@bigfoot.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/*
   This module exports some procedures and functions about arrange
   memory. The main function is pmalloc(). Similar to standard
   malloc(), it's allocate a block of memory by requsition. The 
   distingush is that pmalloc() will call a handle procedure since
   the allocating fails, then all abort. So pmalloc() ought to
   never return a NULL pointer.
   
   pmalloc(size_t, AllocHandle)
   setAllocHandle()		set a default handle
   clearAllocHandle()	clear the default handle
*/

#include	"pmalloc.h"


static AllocHandler	memHandler = NULL;		/* default handle */


void setAllocHandle(AllocHandler handle)
{
	if (memHandler == NULL)  memHandler = handle;
}

void clearAllocHandle()
{
	memHandler = NULL;
}


/* allocate a block of memory. If it fails, clear() will be called
   before exit the program */

void *pmalloc(size_t size, AllocHandler clear)
{
	void    *p;
		
	p = malloc(size);
	if (p == NULL)  {
		
		if (clear == NULL)  {
			if (memHandler)  memHandler();
		} else clear();
		
		exit(-1);
	}
	return p;
}


