
/*
   pop3.c -- the routine set of POP3 protocol.
	
   Copyright (C) 1998, 1999  Xuming <xuming@bigfoot.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<unistd.h>
#include	"defs.h"
#include	"pop3.h"
#include	"pmalloc.h"
#include	"socket.h"


static	POP3	*_pop3 = NULL;

static	int 	_topLists = 0;       
static	int 	_topStyle = SIMPLETOP;
static	int 	_sizeLimit = 0;
static	int 	_loginDelay = 0;



static void setHostname(POP3 *pop, char *s);
static void setId(POP3 *pop, char *s);
static void setIdFromCon(POP3 *pop);
static void setPass(POP3 *pop, char *s);
static void setPassFromCon(POP3 *pop);
static void setMailbox(POP3 *pop, char *s);
static void setAction(POP3 *pop, char *s);
static void destruction(POP3 *pop);

#ifdef	DEBUG
static void exposure(POP3 *pop);
#endif

static POP3 *allocPop3Rc();
static POP3 *AppendingHost();
static void popSetHostFromCon();
static void popSetAltHost();


static int popCheckAuthorization(POP3 *pop);
static int popUser(POP3 *pop);
static int popPass(POP3 *pop);
static int popCheckMailbox(POP3 *pop);
static int popRetr(POP3 *pop);
static int popList(POP3 *pop);
static int popToplist(POP3 *pop);
static int popDelete(POP3 *pop);
static int popQuit(POP3 *pop);
static int pop3Exec(POP3 *pop);

static int popTransact(int socket, char *cmd, int len);



static void setHostname(POP3 *pop, char *s)
{
    pop->hostname = pmalloc(strlen(s)+1, NULL);
    strcpy(pop->hostname, s);
}

static void setId(POP3 *pop, char *s)
{
    if (isblankline(s))  return;
    pop->id = pmalloc(strlen(s)+1, NULL);
    strcpy(pop->id, s);
}

static void setIdFromCon(POP3 *pop)
{
    char	buf[SVRBUF];
    
    if (pop->id == NULL)  {
    	do {
    	    fprintf(stderr, "%s login: ", pop->hostname);
    	    fgets(buf, SVRBUF, stdin);
    	    chop(buf);
    	} while (isblankline(buf));
    	pop->id = pmalloc(strlen(buf)+1, NULL);
    	strcpy(pop->id, buf);
    }
}

static void setPass(POP3 *pop, char *s)
{
    if (!strcmp(s, "*"))  return;
    pop->pass = pmalloc(strlen(s)+1, NULL);
    strcpy(pop->pass, s);
}

static void setPassFromCon(POP3 *pop)
{
    char	buf[SVRBUF];
    
    if (pop->pass == NULL)  {
    	fprintf(stderr, "%s Enter key: ", pop->hostname);
    	strcpy(buf, getpass(""));
    	fprintf(stderr, "\n");
    	pop->pass = pmalloc(strlen(buf)+1, NULL);
    	strcpy(pop->pass, buf);
    }
}


static void setMailbox(POP3 *pop, char *s)
{
    char	buf[SVRBUF], *p;
    
    p = getAltMailbox();
    if (p == NULL)  p = s;
    if (isblankline(p))  p = HOMEBOX;
    getAbsolutePath(p, buf);
    pop->mailbox = pmalloc(strlen(buf)+1, NULL);
    strcpy(pop->mailbox, buf);
}

static void setAction(POP3 *pop, char *s)
{
    pop->action = defaultPop3Action();
    if (pop->action == POP3_NONE)  {
    	switch (*s)  {
    	    case 'c':
    		pop->action = POP3_CK;
    		break;
    	    case 't':
    		pop->action = POP3_TOP;
    		break;
    	    case 'd':
    		pop->action = POP3_DL;
    		break;
    	    case 'f':
    		pop->action = POP3_FETC;
    		break;
    	    default:
    		pop->action = POP3_NONE;
    	}
    }
}

static void destruction(POP3 *pop)
{
    pfree(pop->hostname);
    pfree(pop->id);
    pfree(pop->pass);
    pfree(pop->mailbox);
    pfree(pop->timeStamp);
    pfree(pop->mailList);
    if (pop->socket)  {
    	close(pop->socket);
    	pop->socket = 0;
    }
    pfclose(pop->fp);
}


#ifdef	DEBUG

static void exposure(POP3 *pop)
{
    dbgPrintf("POP3 host name is [%s]\n", pop->hostname);
    dbgPrintf("login identitify is [%s]\n", 
    		pop->id ? pop->id : "NONE");
    dbgPrintf("pass word is [%s]\n", pop->pass ? pop->pass : "NONE" );
    dbgPrintf("mail box is set to [%s]\n",
    		pop->mailbox ? pop->mailbox : "NONE" );
    dbgPrintf("action after login [%d]\n", pop->action);
}

#endif
    


/************************************************************/

static POP3 *allocPop3Rc()
{
    POP3	*ps;
    
    ps = pmalloc(sizeof(POP3), NULL);
    
    ps->hostname	= NULL;
    ps->id		= NULL;
    ps->pass		= NULL;
    ps->mailbox		= NULL;
    ps->action		= POP3_DL;
    ps->next		= NULL;
    ps->socket		= 0;
    ps->fp		= NULL;
    
    ps->timeStamp	= NULL;
    ps->mailList	= NULL;
    ps->mailnum		= 0;
    ps->mailLen		= 0;
    
    return ps;
}

static POP3 *AppendingHost()
{
    POP3	*findLast;
    
    if (_pop3 == NULL)  findLast = _pop3 = allocPop3Rc();
    else {
    	findLast = _pop3;
    	while ( findLast->next )  findLast = findLast->next;
    	findLast->next = allocPop3Rc();
    	findLast = findLast->next;
    }
    return findLast;
}



/********************************************************/


static int popCheckAuthorization(POP3 *pop)
{
    char	buf[SVRBUF], *p;
    int		rs;
    
    rs = readLimitedLine(pop->socket, buf, SVRBUF, timeout());
    if (rs != OK)  return rs;
    
    if (isVerbose())  conPrintf("<< %s", buf);
    if (*buf != '+')  {
    	logPrintf("pop3: what answered? %s", buf);
    	rs = popQuit(pop);
    	if (rs != OK)  return rs;
    	return MISANSWER;
    }
    
    if ((p = strrchr(buf, '>')) == NULL)  return OK;
    *p = '\0';
    if ((p = strchr(buf, '<')) == NULL)  return OK;
    p++;
    pop->timeStamp = pmalloc(strlen(p)+1, NULL);
    strcpy(pop->timeStamp, p);
    
#ifdef	DEBUG
    dbgPrintf("get time stamp from server is %s\n", p);
#endif

    return OK;
}

static int popUser(POP3 *pop)
{
    char	buf[SVRBUF];
    
    sprintf(buf, "USER %s", pop->id);
    return popTransact(pop->socket, buf, SVRBUF);
}
    
static int popPass(POP3 *pop)
{
    char	buf[SVRBUF];
    
    sprintf(buf, "PASS %s", pop->pass);
    return  popTransact(pop->socket, buf, SVRBUF);
}
    
static int popCheckMailbox(POP3 *pop)
{
    char	buf[SVRBUF], *p;
    int 	rs, i;
    
    
    /* use STAT command to check the mailbox. Get the mails
       number and the total size of them */
    
    sprintf(buf, "STAT");
    rs = popTransact(pop->socket, buf, SVRBUF);
    if (rs != OK)  return rs;
    	
    strtok(buf, " ");
    p = strtok(NULL, " ");
    pop->mailnum = atoi(p);
    p = strtok(NULL, " ");
    pop->mailLen = atoi(p);
    if (pop->mailnum == 0)  return OK;
    
    
    /* if there were mails existed, we use LIST command to get
       each letter's size */
    
    pop->mailList = pmalloc(pop->mailnum * sizeof(int), NULL);
    
    sprintf(buf, "LIST");
    rs = popTransact(pop->socket, buf, SVRBUF);
    if (rs != OK)  return rs;

    for (i = 0; i < pop->mailnum; i++)  {
    	rs = readLimitedLine(pop->socket, buf, SVRBUF, timeout());
    	if (rs != OK)  return rs;
    	if (!strcmp(buf, ".\n"))  break;
    	if (isVerbose())  conPrintf("<< %s", buf);
    	strtok(buf, " ");
    	p = strtok(NULL, " ");
    	pop->mailList[i] = atoi(p);
    }
    return readLimitedLine(pop->socket, buf, SVRBUF, timeout());
}
    	
static int popRetr(POP3 *pop)
{
    char	buf[SVRBUF];
    int 	rs, i;
    
    pop->fp = fopen(pop->mailbox, "at");
    if (pop->fp == NULL)  return OPENFILERR;
    
    for (i = 1; i <= pop->mailnum; i++)  {
    	if ( !isInSet(i) )  continue;
    	
    	conPrintf("get mail %3d/%d (%7d) ", 
    		i, pop->mailnum, pop->mailList[i-1]);
    	
    	if (_sizeLimit && (_sizeLimit < pop->mailList[i-1]))  {
    	    conPrintf("oversized (%dK limited), skip it.\n",
    			_sizeLimit / 1024);
    	    continue;
    	}
    	
    	if (isVerbose())  conPrintf("\n");
    	sprintf(buf, "RETR %d", i);
    	rs = popTransact(pop->socket, buf, SVRBUF);
    	if (rs != OK)  {
    	    pfclose(pop->fp);
    	    return rs;
    	}
    	
    	fprintf(pop->fp, "From - %s\n", timenow());
    	dotListInit(pop->mailList[i-1], 40);
    	while(1)  {
    	    rs = socketReadLine(pop->socket, buf, SVRBUF, timeout());
    	    if (rs != OK)  {
    		pfclose(pop->fp);
    		return rs;
    	    }
    	    if ( !strcmp(buf, ".\n") )  break;
    		
    	    dotIllustrate(strlen(buf));
    	    if (*buf == '.')  fputs(buf+1, pop->fp);
    	    else  fputs(buf, pop->fp);
    	}
    	fputs("\n", pop->fp);
    	dotToEnd();
    	
    	if (pop->action == POP3_DL)  {
    	    sprintf(buf, "DELE %d", i);
    	    rs = popTransact(pop->socket, buf, SVRBUF);
    	    if (rs != OK)  {
    		pfclose(pop->fp);
    		return rs;
    	    }
    	}
    }
    pfclose(pop->fp);
    return OK;
}

static int popList(POP3 *pop)
{
    int 	i;
    
    for (i = 0; i < pop->mailnum; i++)  {
    	if (isInSet(i+1))  {
    	    conPrintf("%3d# of %d, %7d bytes\n",  
    			i+1, pop->mailnum, pop->mailList[i]);
    	}
    }
    return OK;
}

static int popToplist(POP3 *pop)
{
    char	buf[SVRBUF];
    int 	rs, i, isHeader;
    
    for (i = 1; i <= pop->mailnum; i++)  {
    	if ( !isInSet(i) )  continue;
    	
    	sprintf(buf, "TOP %d %d", i, _topLists);
    	rs = popTransact(pop->socket, buf, SVRBUF);
    	if (rs == MISANSWER)  {
    	    conPrintf("TOP command is't support. Try list instead\n");
    	    popList(pop);
    	    break;
    	} if (rs != OK)  return rs;
    	
    	conPrintf("Mail %d# of %d, %d bytes\n", 
    			i, pop->mailnum, pop->mailList[i-1]);
    	
    	isHeader = 1;
    	while(1)  {
    	    rs = socketReadLine(pop->socket, buf, SVRBUF, timeout());
    	    if (rs != OK)  return rs;
    	    if ( !strcmp(buf, ".\n") )  break;
    	    if (isblankline(buf))  isHeader = 0;
    	
    	    switch(_topStyle)  {
    		case SIMPLETOP:
    		    if (!isHeader)  break;
    		    if (!strnicmp(buf, "From:", 5) ||
    			!strnicmp(buf, "Return-Path:", 12) ||
    			!strnicmp(buf, "Sender:", 7))  {
    					
    			conPrintf(buf);
    		    }
    		    break;
    				
    		case COMPLEXTOP:
    		    conPrintf(buf);
    		    break;
    				
    		case PARTBODY:
    		    if (!isHeader)  conPrintf(buf);
    		    break;
    				
    		case SMARTOP:
    		    if (!strnicmp(buf, "Content-", 8))  break;
    		    if (!strnicmp(buf, "Message-Id:", 11))  break;
    		    if (!strnicmp(buf, "X-", 2))  break;
    		    if (!strnicmp(buf, "MIME", 4))  break;
    		    if (!strnicmp(buf, "References:", 11))  break;
    		    if (!strnicmp(buf, "Status:", 7))  break;
    		    if (!strnicmp(buf, "Received:", 9)) break;
    		    if ((*buf <= ' ') && !isblankline(buf))  break;
    		    conPrintf(buf);
    	    }
    	}
    	
    	conPrintf("\n");
    }
    return OK;
}

static int popDelete(POP3 *pop)
{
    char	buf[SVRBUF];
    int 	rs, counter, i;
    
    counter = 0;
    for (i = 1; i <= pop->mailnum; i++)  {
    	if ( !isSetExist() || !isInSet(i))  continue;

    	conPrintf("remove mail %3d of %d, %d bytes\n", 
    		i, pop->mailnum, pop->mailList[i-1]);
    	
    	sprintf(buf, "DELE %d", i);
    	rs = popTransact(pop->socket, buf, SVRBUF);
    	if (rs != OK)  return rs;
    	
    	counter ++;
    }
    conPrintf("Remove %d mail(s) from %s directly\n", 
    		counter, pop->hostname);
    return OK;
}

static int popQuit(POP3 *pop)
{
    char	buf[SVRBUF];
    int 	rs;

    if (isVerbose())  conPrintf("-> QUIT\n");
    sprintf(buf, "QUIT\r\n");
    
    rs = socketWrite(pop->socket, buf, strlen(buf), timeout());
    if (rs != OK)  return rs;
    
    rs = readLimitedLine(pop->socket, buf, SVRBUF, timeout());
    if (rs != OK)  return rs;
    if (isVerbose() || (*buf != '+'))  conPrintf("<< %s", buf);
    
    return OK;
}
    

static int pop3Exec(POP3 *pop)
{
    int		rs;
    
    setIdFromCon(pop);
    setPassFromCon(pop);
    
    conPrintf("Trying POP3 host %s as %s ...\n", 
    		pop->hostname, pop->id);
    rs = OpenSocket(pop->hostname, POP3_PORT);
    if (rs < 0)  return rs;
    pop->socket = rs;
    
    if ((rs = popCheckAuthorization(pop)) != OK)  {
    	destruction(pop);
    	return rs;
    }
    
    if ((rs = popUser(pop)) != OK)  {
    	conPrintf("User %s is rejected.\n", pop->id);
    	destruction(pop);
    	return rs;
    }
    
    if ((rs = popPass(pop)) != OK)  {
    	conPrintf("User %s certificate failed.\n", pop->id);
    	destruction(pop);
    	return rs;
    }
    
    if (_loginDelay)  sleep(_loginDelay);
    logPrintf("pop3: login as %s into %s\n", pop->id, pop->hostname);
    
    if ((rs = popCheckMailbox(pop)) != OK)  {
    	destruction(pop);
    	return rs;
    }
    
    if ( pop->mailnum )  {
    	conPrintf("%d mail(s) are ready, total %d bytes\n", 
    			pop->mailnum, pop->mailLen);
    } else {
    	conPrintf("No mail is ready in %s\n", pop->hostname);
    	popQuit(pop);
    	destruction(pop);
    	return OK;
    }
    
    switch (pop->action) {
    	case POP3_DL:
    	case POP3_FETC:
    	    if ((rs = popRetr(pop)) != OK)  {
    	 	destruction(pop);
    		return rs;
    	    }
    	    logPrintf("pop3: download %d mails from %s.\n", 
    			pop->mailnum, pop->hostname);
    	    break;
    		
    	case POP3_CK:
    	    if ((rs = popList(pop)) != OK)  {
    		destruction(pop);
    		return rs;
    	    }
    	    logPrintf("pop3: %d mails is listed.\n", 
   			pop->mailnum);
    	    break;
    		
    	case POP3_TOP:
    	    if ((rs = popToplist(pop)) != OK)  {
    		destruction(pop);
    		return rs;
    	    }
    	    logPrintf("pop3: fetch %d mails part body.\n", 
    			pop->mailnum);
    	    break;
    		
    	case POP3_RM:
    	    if ((rs = popDelete(pop)) != OK)  {
    		destruction(pop);
    		return rs;
    	    }
    	    logPrintf("pop3: %d mails was especially removed.\n",
    			pop->mailnum);
    	    break;
    		
    	default:
    	    logPrintf("pop3: connect to host but do nothing.\n");
    	    break;
    }
    popQuit(pop);
    destruction(pop);
    return OK;
}

    
static void popSetHostFromCon()
{
    POP3	*ps;
    char	*s, buf[SVRBUF];

    ps = AppendingHost();
    
    if ((s = getAltHost()) == NULL)  {
    	do {
    	    fprintf(stderr, "The POP3 mail server is: ");
    	    fgets(buf, SVRBUF, stdin);
    	    chop(buf);
    	} while (isblankline(buf));
    	s = buf;
    }
    setHostname(ps, s);
    
    if ((s = getAltMailbox()) == NULL)  {
    	fprintf(stderr, "Local mailbox is (%s): ", HOMEBOX);
    	fgets(buf, SVRBUF, stdin);
    	chop(buf);
    	if (isblankline(buf))  strcpy(buf, HOMEBOX);
    	s = buf;
    }
    setMailbox(ps, s);
    
    ps->action = defaultPop3Action();
}
    	
static void popSetAltHost()
{
    char	*p, buf[SVRBUF];
    POP3	*ps;
    
    if (_pop3 == NULL)  {
    	ps = AppendingHost();
    	setHostname(ps, getAltHost());
    	ps->action = defaultPop3Action();
    		
    	if ((p = getAltMailbox()) == NULL)  {
    		/*
    		fprintf(stderr, "Local mailbox is (%s): ", HOMEBOX);
    		fgets(buf, SVRBUF, stdin);
    		chop(buf);
    		if (isblankline(buf))  */
    	    strcpy(buf, HOMEBOX);
    	    p = buf;
    		
    	}
    	setMailbox(ps, p);
    }
}


/*************************************************************/

static int popTransact(int socket, char *cmd, int len)
{
    char	buf[SVRBUF];
    int 	rs;
    
    if (isVerbose())  conPrintf("-> %s\n", cmd);
    sprintf(buf, "%s\r\n", cmd);
    rs = socketWrite(socket, buf, strlen(buf), timeout());
    if (rs != OK)  return rs;
    
    rs = readLimitedLine(socket, buf, SVRBUF, timeout());
    if (rs != OK)  return rs;	
    if (isVerbose())  conPrintf("<< %s", buf);
    if (*buf != '+')  {
    	logPrintf("pop3 abort: %s failed: %s", cmd, buf);
    	strcpy(buf, "QUIT\r\n");
    	rs = socketWrite(socket, buf, strlen(buf), timeout());
    	if (rs != OK)  return rs;
    	rs = readLimitedLine(socket, buf, SVRBUF, timeout());
    	if (rs != OK)  return rs;
    	return MISANSWER;
    }
    strncpy(cmd, buf, len);
    return OK;
}	
    
/**********************************************************/

void popSetHost(char *s)
{
    char	buf[SVRBUF], *p, *delim;
    POP3	*ps;

    if ((p = getAltHost()) && !isblankline(p))  {
    	popSetAltHost();
    	return;
    }
    
    /* because of using run-time function strtok() in the procedure
       getAbsolutePath(), I have to parse data directly avoiding 
       reenterment strtok()  */
    
    strcpy(buf, s);
    p = buf;
    if ((delim = strchr(buf, ':')) == NULL)  return;
    *delim++ = '\0';
    if (isblankline(p))  return;		/* host name must available */	
    ps = AppendingHost();
    setHostname(ps, p);
    
    p = delim;
    if ((delim = strchr(delim, ':')) == NULL)  return;
    *delim++ = '\0';
    setAction(ps, p);
    
    p = delim;
    if ((delim = strchr(delim, ':')) == NULL)  return;
    *delim++ = '\0';
    setMailbox(ps, p);
    
    p = delim;
    if ((delim = strchr(delim, ':')) == NULL)  return;
    *delim++ = '\0';
    setId(ps, p);

    setPass(ps, delim);
}


void popListTopLines(char *s)
{
    _topLists = atoi(s);
}

void popTopMethod(char *s)
{
    if (!strcmp(s, DEF_SIMPLE))	 _topStyle = SIMPLETOP;
    else if (!strcmp(s, DEF_COMPLEX))  _topStyle = COMPLEXTOP;
    else if (!strcmp(s, DEF_SMART))  _topStyle = SMARTOP;
}

void popSizeLimitation(char *s)
{
    _sizeLimit = atoi(s);
    s = GotoLineEnd(s);
    switch (*s)  {
    	case 'k':
    	case 'K':
    	    _sizeLimit *= 1024;
    	    break;
    	case 'm':
    	case 'M':
    	    _sizeLimit *= 1024 * 1024;
    	    break;
    }
    if (_sizeLimit < 0)  _sizeLimit = 0;
}

void popLoginDelay(char *s)
{
    _loginDelay = atoi(s);
}



#ifdef	DEBUG

void popVarietyExposure()
{
    POP3	*pop;
    
    dbgPrintf("List top [%d] lines of mail body\n", _topLists);
    dbgPrintf("User top style is [%d]\n", _topStyle);
    dbgPrintf("mail size is limited in [%d]\n", _sizeLimit);
    dbgPrintf("delay time when login on is [%d]\n", _loginDelay);
    
    for (pop = _pop3; pop; pop = pop->next)  exposure(pop);
}

#endif
    

/***********************************************************/

void pop3main()
{
    POP3	*pop;
    int 	rs;
    
    if (_pop3 == NULL)  popSetHostFromCon();
    for (pop = _pop3; pop != NULL; pop = pop->next)  {
    	rs = pop3Exec(pop);
    	if (rs != OK)  errReport(rs, "pop3");
    	conPrintf("\n");
    }
    popReleaseAll();
}



void popReleaseAll()
{
    POP3	*ps, *tmp;
    
    ps = _pop3;
    while ( ps )  {
    	tmp = ps,
    	ps = ps->next;
    	destruction(tmp);
    	pfree(tmp);
    }
    _pop3 = NULL;
}

