
/*
   progress.c -- display the rate of progress when download
   or upload something. The TABLE STOP is set to 4 characters 
   in this file.
	
   Copyright (C) 1998, 1999  Xuming <xuming@bigfoot.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/*
   progree.c
   
   This module exports some procedure for indicating the progress
   of download/upload mail by displaying a string of dot in the 
   standard output (file description 1). This can be change in this
   file, which defined as OUTPUTOi.
   
   dotListInit()	every time to draw the progress, this
                    procedure should be called first
   dotIllustrate()  give it increased data length
   dotToEnd()		do not forget it when dot list is over

   Note that the last rate is average transfer rate. See dotToEnd()
   for details.
*/

#include	<stdio.h>
#include	<string.h>
#include	<unistd.h>
#include	<sys/time.h>


#define		MAXINT		((int)(~0U>>1))
#define		OUTPUTTO	1


static	int 	dotted = 0;		/* the number of printed dots */
static	int 	dotall = 0;		/* the expected number for printing */
static	int 	gotten = 0;		/* the byte number has fetched */
static	int 	expect = 0;		/* the estimated length of file */
static	int		lastgot = 0;	/* the byte number last fetched */

static	struct	timeval	start;	/* used for compute average tranfer */
                                /*  rate */
static	struct	timeval	last;


static void transRate(struct timeval *rct, int increased, char *buf);


void dotListInit(int expectLength, int dotLength)
{
	dotted  = 0;
	gotten  = 0;
	lastgot = 0;
	expect  = expectLength;
	dotall  = dotLength;
	gettimeofday(&start, (struct timezone *) NULL);
	last = start;
}

void dotIllustrate(int cur)
{
	int 	dotStop, i;
	char	rate[16];
	
	if ( !expect || !dotall )  return;
	
	/* compute that how many dots should be printed by now */
	
	gotten += cur;
	dotStop = (dotall * gotten) / expect;
	
	if ((dotted < dotStop) && (dotted < dotall))  {
		
		/* print dots from last position to where it should be in */
		
		for (; (dotted < dotStop) && (dotted < dotall); dotted++)  {
			write(OUTPUTTO, ".", 1);
		}
		
		transRate(&last, gotten - lastgot, rate);
		write(OUTPUTTO, rate, strlen(rate));
		
		lastgot = gotten;
		for (i = 0; i < strlen(rate); i++)  write(OUTPUTTO, "\b", 1);
	}
}
	

void dotToEnd()
{
	char	rate[16];
	
	if ( !expect || !dotall )  return;
	
	for (; dotted < dotall; dotted++)  write(OUTPUTTO, ".", 1);
	transRate(&start, gotten, rate);
	write(OUTPUTTO, rate, strlen(rate));
	write(OUTPUTTO, "\n", 1);
	dotted  = 0;
	gotten  = 0;
	lastgot = 0;
	expect  = 0;
	dotall  = 0;
}



/* produce data transfer rate, then store it in buf.
   the rate will be formated of constant length. i.e
   
      " 1024 bps" or " 10.2 K/s"
*/
   
static void transRate(struct timeval *rct, int increased, char *buf)
{
	char	*denomi[] = { "bps", "K/s", "M/s", "G/s" };
	struct	timeval	now, diff;
	int 	i, m, rate, divisor;

	/* if the data stream rate is too large or unexpectation happed,
	   use violent as default infor */
	
	strcpy(buf, "  violent");
	
	
	/* compute the time step increasing. That is to substract last
	   time, which is the first parameter, from current time. The
	   resule stores in variety 'now'. */

	gettimeofday(&now, (struct timezone *) NULL);
	diff.tv_sec  = now.tv_sec - rct->tv_sec;
	diff.tv_usec = now.tv_usec - rct->tv_usec;
	last = now;
	if (diff.tv_usec < 0)  {
		diff.tv_usec += 1000000;
		diff.tv_sec --;
	}
	if (diff.tv_sec < 0)  return;
	if ((diff.tv_sec == 0) && (diff.tv_usec == 0))  return;
	
	
	rate = 0;
	if (increased >= diff.tv_sec)  {
		int 	uplimit;
		
		for (i = 1000000; i > 0; i = i / 10)  {
			uplimit = MAXINT / i;
			if (increased < uplimit)  {
				divisor = (diff.tv_sec * i) + 
						diff.tv_usec / (1000000 / i);
				if (divisor > 0)  rate = (increased * i) / divisor;
				break;
			}
		}
	/*	printf("incr is %d, i is %d, divisor is %d, time %d:%d\n", 
				increased, i, divisor, diff.tv_sec, diff.tv_usec);*/
	}
	
	/* append the denomination */
	
	for (m = 0, i = 0; i < 4; i++)  {
		if (rate < 1000)  break;
		m = rate % 1000;
		rate = rate / 1000;
	}
   	if (i == 0)  {
		sprintf(buf, " %4d %s", rate, denomi[0]);
	} else  if (rate < 10)  {
		sprintf(buf, " %d%03d %s", rate, m, denomi[i-1]);
	} else  if (rate < 100) {
		sprintf(buf, " %d.%d %s", rate, m/100, denomi[i]);
	} else {
		sprintf(buf, " %4d %s", rate, denomi[i]);
	}
}

