/*
   smtp.c -- the routine set of SMTP
   The TABLE STOP is set to 4 characters in this file.
	
   Copyright (C) 1998, 1999  Xuming <xuming@bigfoot.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include	<stdio.h>
#include	<string.h>
#include	<unistd.h>
#include	<pwd.h>
#include	"defs.h"
#include	"pmalloc.h"
#include	"smtp.h"
#include	"socket.h"


static	char	*_smtpHost = NULL;
static	char	*_outbox = NULL;
static	char	*_backupbox = NULL;
static	char	*_signature = NULL;
static	char	*_pipeto = NULL;
static	char	*_sender = NULL;
static	int 	_sentStrategy = SMTP_BACKUP;

static	char	*_subject = NULL;	/* from command line */
static	char	*_cc = NULL;
static	char	*_bcc = NULL;

static	int 	_enforceSendmail = 0;

static	char	*_extraHeader[EXTRAHEADER];
static	int 	_headerNum = 0;
static	int 	_headerCnt = 0;

static	char	**_toaddrList = NULL;
static	int 	_toaddrCnt = 0;

static	int 	_sock = 0;


static void smtpCheckParam();
static int smtpHelo(int sock);
static int smtpMailFrom(int sock, char *cmd);
static int smtpRcptTo(int sock, char *cmd);
static int smtpData(int sock);
static int smtpDotClose(int sock);
static int smtpRset(int sock);
static int smtpQuit(int sock);
static int smtpAnswer(int sock, char *buf, int len);
static int smtpTransact(int sock, char *cmd, int len);


static void smtpCheckParam()
{
	char	*p, buf[SVRBUF];
	
	if (_smtpHost == NULL)  {
		if ((p = getAltHost()) != NULL)  strcpy(buf, p);
		else  do {
			fprintf(stderr, "The SMTP server: ");
			fgets(buf, SVRBUF, stdin);
			chop(buf);
		} while (isblankline(buf));
		_smtpHost = pmalloc(strlen(buf)+1, NULL);
		strcpy(_smtpHost, buf);
	}
	
	if (_outbox == NULL)  {
		char	mb[SVRBUF];
		
		if ((p = getAltMailbox()) && _enforceSendmail)  strcpy(buf, p);
		else  {
			fprintf(stderr, "The outgoing mailbox: ");
			fgets(buf, SVRBUF, stdin);
			chop(buf);
		}
		getAbsolutePath(buf, mb);
		_outbox = pmalloc(strlen(mb)+1, NULL);
		strcpy(_outbox, mb);
	}
	
	if (_sender == NULL)  {
		fprintf(stderr, "Reply to: ");
		fgets(buf, SVRBUF, stdin);
		chop(buf);
		if (!isblankline(buf))  {
			_sender = pmalloc(strlen(buf)+1, NULL);
			strcpy(_sender, buf);
		}
	}
}		
	
	

static int smtpHelo(int sock)
{
	char	buf[SVRBUF];
	int 	rs;
	struct	passwd	*pw;
	
	pw = getpwuid(geteuid());
	sprintf(buf, "HELO %s", pw->pw_name);
	rs = smtpTransact(sock, buf, SVRBUF);
	if (rs != OK)  return rs;
	if (!strncmp(buf, "250", 3))  return OK;
	return MISANSWER;
}

static int smtpMailFrom(int sock, char *cmd)
{
	char	buf[SVRBUF];
	int 	rs;
	
	if (cmd == NULL)  strcpy(buf, "MAIL FROM:<>");
	else  sprintf(buf, "MAIL FROM:<%s>", emailAddr(cmd));
	rs = smtpTransact(sock, buf, SVRBUF);
	if (rs != OK)  return rs;
	
	if (!strncmp(buf, "250", 3))  return OK;
	
	strcpy(buf, "QUIT");
	rs = smtpTransact(sock, buf, SVRBUF);
	if (rs != OK)  return rs;
	return MISANSWER;
}

static int smtpRcptTo(int sock, char *cmd)
{
	char	buf[SVRBUF];
	int 	rs;
	
	if (cmd == NULL)  return OK;
	sprintf(buf, "RCPT TO:<%s>", emailAddr(cmd));
	rs = smtpTransact(sock, buf, SVRBUF);
	if (rs != OK)  return rs;
	
	if (strncmp(buf, "250", 3))  {
		logPrintf("smtp: mail to %s, %s", cmd, buf);
	}
	return OK;
}

static int smtpData(int sock)
{
	char	buf[SVRBUF];
	int 	rs;
	
	strcpy(buf, "DATA");
	rs = smtpTransact(sock, buf, SVRBUF);
	if (rs != OK)  return rs;
	
	if (!strncmp(buf, "354", 3))  return OK;
	return MISANSWER;
}

static int smtpDotClose(int sock)
{
	char	buf[SVRBUF];
	int 	rs;
	
	strcpy(buf, ".");
	rs = smtpTransact(sock, buf, SVRBUF);
	if (rs != OK)  return rs;
	
	if (!strncmp(buf, "250", 3))  return OK;
	
	strcpy(buf, "QUIT");
	rs = smtpTransact(sock, buf, SVRBUF);
	if (rs != OK)  return rs;
	return MISANSWER;
}

static int smtpRset(int sock)
{
	char	buf[SVRBUF];
	int 	rs;
	
	strcpy(buf, "RSET");
	rs = smtpTransact(sock, buf, SVRBUF);
	if (rs != OK)  return rs;
	return OK;
}

static int smtpQuit(int sock)
{
	char	buf[SVRBUF];
	int 	rs;
	
	strcpy(buf, "QUIT");
	rs = smtpTransact(sock, buf, SVRBUF);
	if (rs != OK)  return rs;
	return OK;
}

static int smtpAnswer(int sock, char *buf, int len)
{
	char	tmp[SVRBUF];
	int 	rs;
	
	rs = readLimitedLine(sock, buf, len, timeout());
	if (rs != OK)  return rs;
	if (isVerbose())  conPrintf("<< %s", buf);
	
	strcpy(tmp, buf);
	while (tmp[3] == '-')  {
		rs = readLimitedLine(sock, tmp, SVRBUF, timeout());
		if (rs != OK)  return rs;
		if (isVerbose())  conPrintf("<< %s", tmp);
	}
	return OK;
}

static int smtpTransact(int sock, char *cmd, int len)
{
	char    buf[SVRBUF];
	int 	rs;
	
	if (isVerbose())  conPrintf("-> %s\n", cmd);
	sprintf(buf, "%s\r\n", cmd);
	rs = socketWrite(sock, buf, strlen(buf), timeout());
	if (rs != OK)  return rs;
	rs = smtpAnswer(sock, cmd, len);
	return rs;
}
		

/**********************************************************/


void smtpEnforced()
{
	_enforceSendmail = 1;
}
	
void smtpSetSubject(char *s)
{
	pfree(_subject);
	_subject = pmalloc(strlen(s)+1, NULL);
	strcpy(_subject, s);
}

char *smtpGetSubject()
{
	return _subject;
}

void smtpSetCc(char *s)
{
	pfree(_cc);
	_cc = pmalloc(strlen(s)+1, NULL);
	strcpy(_cc, s);
}

char *smtpGetCc()
{
	return _cc;
}

void smtpSetBcc(char *s)
{
	pfree(_bcc);
	_bcc = pmalloc(strlen(s)+1, NULL);
	strcpy(_bcc, s);
}

char *smtpGetBcc()
{
	return _bcc;
}

void smtpSetToaddr(int cnt, char **list)
{
	_toaddrCnt = cnt;
	_toaddrList = list;
}

char **smtpGetToaddr(int *cnt)
{
	*cnt	= _toaddrCnt;
	return _toaddrList;
}

void smtpSetHost(char *s)
{
	char	*ah;
	
	ah = getAltHost();
	if (ah && !isblankline(ah))  s = ah;
	
	pfree(_smtpHost);
	_smtpHost = pmalloc(strlen(s)+1, NULL);
	strcpy(_smtpHost, s);
}

void smtpSetSender(char *s)
{
	pfree(_sender);
	_sender = pmalloc(strlen(s)+1, NULL);
	strcpy(_sender, s);
}

char *smtpGetSender()
{
	return _sender;
}

void smtpSetOutbox(char *s)
{
	char	*am, buf[SVRBUF];
	
	am = getAltMailbox();
	if (am && _enforceSendmail)  s = am;
	getAbsolutePath(s, buf);
		
	pfree(_outbox);
	_outbox = pmalloc(strlen(buf)+1, NULL);
	strcpy(_outbox, buf);
}

char *smtpGetOutbox()
{
	return _outbox;
}

void smtpSetBackupbox(char *s)
{
	char	buf[SVRBUF];
	
	getAbsolutePath(s, buf);
	pfree(_backupbox);
	_backupbox = pmalloc(strlen(buf)+1, NULL);
	strcpy(_backupbox, buf);
}

char *smtpGetBackupbox()
{
	return _backupbox;
}

void smtpSetPipeto(char *s)
{
	pfree(_pipeto);
	_pipeto = pmalloc(strlen(s)+1, NULL);
	strcpy(_pipeto, s);
}

char *smtpGetPipeto()
{
	return _pipeto;
}

void smtpSetDowithSent(char *s)
{
	if (!strcmp(s, DEF_HOLDON))			_sentStrategy = SMTP_HOLDON;
	else  if (!strcmp(s, DEF_DELETE))	_sentStrategy = SMTP_DELETE;
	else  if (!strcmp(s, DEF_BACKUP))	_sentStrategy = SMTP_BACKUP;
}

int smtpGetDowithSent()
{
	return _sentStrategy;
}

void smtpSetSignature(char *s)
{
	char	buf[SVRBUF];
	
	getAbsolutePath(s, buf);
	pfree(_signature);
	_signature = pmalloc(strlen(buf)+1, NULL);
	strcpy(_signature, buf);
}

char *smtpGetSignature()
{
	return _signature;
}

void resetAppendixHeader()
{
	_headerCnt = 0;
}

char *getAppendixHeader()
{
	char	*p = NULL;
	
	if (_headerCnt < _headerNum)  {
		p = _extraHeader[_headerCnt];
		_headerCnt++;
	}
	return p;
}

void smtpSetExtraHeader(char *s)
{
	if (_headerNum < EXTRAHEADER)  {
		_extraHeader[_headerNum] = pmalloc(strlen(s)+1, NULL);
		strcpy(_extraHeader[_headerNum], s);
		_headerNum++;
	}
}

void smtpRelease()
{
	int 	i;
	
	closeFolder();
	
	pfree(_smtpHost);
	pfree(_outbox);
	pfree(_backupbox);
	pfree(_signature);
	pfree(_pipeto);
	pfree(_sender);
	pfree(_subject);
	pfree(_cc);
	pfree(_bcc);
	
	if (_sock)  {
		close(_sock);
		_sock = 0;
	}
	
	for (i = 0; i < _headerNum; i++)  pfree(_extraHeader[i]);
	_headerNum = _headerCnt = 0;
}


#ifdef	DEBUG

void smtpVarietyExposure()
{
	int 	i;
	
	dbgPrintf("the smtp host is [%s]\n", _smtpHost);
	dbgPrintf("the outgoing mailbox is [%s]\n", _outbox);
	dbgPrintf("the backup mailbox is [%s]\n", _backupbox);
	dbgPrintf("the signature file is [%s]\n", _signature);
	dbgPrintf("the pipe option is [%s]\n", _pipeto);
	dbgPrintf("the sender (reply-to) is [%s]\n", _sender);
	dbgPrintf("the subject is [%s]\n", _subject);
	dbgPrintf("the cc is [%s]\n", _cc);
	dbgPrintf("the bcc is [%s]\n", _bcc);
	dbgPrintf("the sent mail strategy is %d\n", _sentStrategy);
	
	for (i = 0; i < _headerNum; i++)  {
		dbgPrintf("extra header: %s\n", _extraHeader[i]);
	}
	
	dbgPrintf("the to address is listed %d\n", _toaddrCnt);
	for (i = 0; i < _toaddrCnt; i++)  {
		dbgPrintf("to address list: %s\n", _toaddrList[i]);
	}
}

#endif

/***********************************************************/

int smtpMain(int todo)
{
	char	buf[SVRBUF];
	int 	rs, mailnum = 1, i;
	
	if (isatty(0))  smtpCheckParam();
	else  {
		char *p = getAltHost();
		if (p)  {
			_smtpHost = pmalloc(strlen(p)+1, NULL);
			strcpy(_smtpHost, p);
		}
		if (_smtpHost == NULL)  return NOSMTPHOST;
	}
	
	if (todo == SENDMAIL)  rs = openFolder(_outbox, &mailnum);
	else if (todo == STREAM)  rs = createFolderFromCon();
	else return OK;
	
	if (rs != OK)  return rs;
	if (mailnum == 0)  {
		printf("no mail is ready for sending.\n");
		return OK;
	}
	
	printf("%d mail(s) are ready for sending.\n", mailnum);
	printf("Trying SMTP host [%s] ...\n", _smtpHost);
	
	rs = OpenSocket(_smtpHost, SMTP_PORT);
	if (rs < 0)  return rs;
	_sock = rs;
	logPrintf("smtp: %s connected for sending mails.\n", _smtpHost);
	
   	rs = smtpAnswer(_sock, buf, SVRBUF);
	if (rs != OK)  return rs;
	if (strncmp(buf, "220", 3))  return MISANSWER;
	
	rs = smtpHelo(_sock);
	if (rs != OK)  return rs;
	
	for (i = 0; i < mailnum; i++)  {
		
		char	*tmp, *bulk;
		SENDTO	*st;
		COPYTO	*ct;
		
		st = setActiveMail(i+1);
		if (st == NULL)  break;
		
		conPrintf("send mail %3d/%d, (%-7ld) ", 
				i + 1, mailnum, st->mailLen);
		if (isVerbose())  conPrintf("\n");
		
		if ((rs = smtpRset(_sock)) != OK)  break;
		
		if ((rs = smtpMailFrom(_sock, st->from)) != OK)  {
			conPrintf("invalid mail from address.\n");
			continue;
		}
		
		for (ct = st->to; ct; ct = ct->next)  {
			rs = smtpRcptTo(_sock, ct->host);
			if (rs != OK)  break;
		}
		
		if ((rs = smtpData(_sock)) == MISANSWER)  {
			conPrintf("fade!\n");
			continue;
		} else if (rs != OK)  break;
		
		*buf = '.';
		bulk = buf + 1;
		dotListInit(st->mailLen, 40);
		while (getMail(bulk, SVRBUF - 2))  {
			tmp = GotoLineEnd(bulk);
			if (*tmp == '\n')  {
				*tmp++ = '\r', *tmp++ = '\n', *tmp = '\0';
			}
			if (*bulk == '.')  {
				dotIllustrate(strlen(buf));
				rs = linesend(_sock, buf, timeout());
			} else {
				dotIllustrate(strlen(bulk));
				rs = linesend(_sock, bulk, timeout());
			}
			if (rs != OK)  break;
		}
		if (rs != OK)  break;
		dotToEnd();
		rs = sflush(_sock, timeout());
		if (rs != OK)  break;
		
		rs = smtpDotClose(_sock);
		if (rs != OK)  break;
		
		st->status = _sentStrategy;
	}
	if (rs != OK)  return rs;
	smtpQuit(_sock);
	logPrintf("smtp: %d mails delivered.\n", mailnum);
	
	smtpRelease();
	return OK;
}


