/*
   socket.c -- define some lowlevel function about socket access.
   The TABLE STOP is set to 4 characters in this file.
	
   Copyright (C) 1998, 1999  Xuming <xuming@bigfoot.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/



#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<unistd.h>
#include	<netdb.h>
#include	<sys/types.h>
#include	<sys/socket.h>
#include	<netinet/in.h>
#include	<arpa/inet.h>
#include	<errno.h>
#include	<time.h>
#include	"socket.h"



/*
	OpenSocket --
   
		Create a socket, and make a TCP connection in specified 
	"port" of the "host". 
		The return value is a discription of the socket, or < 0 
	when it fails.
*/ 
int  OpenSocket(char *host, int port)	
{
	struct	sockaddr_in		sa;
	struct	hostent			*hp;
	int 	s;

	bzero(&sa, sizeof(sa));
	sa.sin_family = AF_INET;
	if ((sa.sin_addr.s_addr = inet_addr(host)) == INADDR_NONE)  {		
		if ((hp = gethostbyname(host)) == NULL)  return  MISHOSTS;
		bcopy(hp->h_addr, (char*) &sa.sin_addr, hp->h_length);
		sa.sin_family = hp->h_addrtype;
	}
	if ((s = socket(sa.sin_family, SOCK_STREAM, 0)) < 0)  {
		return MISSOCKET;
	}
	sa.sin_port	= htons(port);
	if (connect(s, (struct sockaddr*)&sa, sizeof(sa)) < 0)  {
		close(s);
		return MISCONNECT;
	}
	return (s);
}


int  socketWrite(int socket, char *buf, int len, int timer)
{
    int 	n;
	
	alarm(timer);
	while (len)  {
		n = write(socket, buf, len);
		if (n <= 0)  {
			alarm(0);
			return LOOKERRNO;
		}
		len -= n;
		buf += n;
	}
	alarm(0);
	return OK;
}


int socketReadLine(int socket, char *buf, int len, int timer)
{
	char	*nl, *p;
	int 	num;
	
	alarm(timer);
	buf[--len] = '\0';
	p = buf;
	do {
		num = recv(socket, p, len, MSG_PEEK);
		if (num <= 0)  {
			alarm(0);
			return MISCONNECT;
		}
		
		nl = memchr(p, '\n', num);
		if (nl)  num = nl - p + 1;
		
		num = read(socket, p, num);
		if (num == -1)  {
			alarm(0);
			return MISCONNECT;
		}
		
		p += num;
		len -= num;
	} while ( !nl && len );
	
	if (nl)  {
		*(nl+1) = '\0';
		if ( *(nl-1) == '\r' )  {
			*(nl-1) = '\n';
			*nl = '\0';
		}
	}
	alarm(0);
	return OK;
}
	

int readLimitedLine(int socket, char *buf, int len, int timer)
{
	char	tmp[SOCKGENBUF];
	int 	rs;
	
	rs = socketReadLine(socket, buf, len, timer);
	strcpy(tmp, buf);
	while ( !isEntireLine(tmp) )  {
		if (socketReadLine(socket, tmp, SOCKGENBUF, timer) <= 0)  
			return MISCONNECT;
	}
	return rs;
}


static	char	netbuf[NETBUF];
static	int		nbidx = 0;

/*
   linesend() sflush() --
   
   linesend() is used to buffer the ready-send data.
   sflush() is very like the fflush() function, to oblige the program
   sent the data in the buffer at once
*/ 
int  linesend(int socket, char *buf, int timer)
{
	int 	c, r = OK;
	
	if (strlen(buf) > NETBUF)  {
		sflush(socket, timer);
		r = socketWrite(socket, buf, strlen(buf), timer);
	} else { 
		if ((nbidx + strlen(buf)) > NETBUF)  r = sflush(socket, timer);
		for (c=0; (nbidx < NETBUF) && buf[c]; nbidx++, c++)  {
			netbuf[nbidx] = buf[c];
		}
	}
	return r;
}

int  sflush(int socket, int timer)
{
	int 	r = OK;
	
	if (nbidx != 0)  {
		r = socketWrite(socket, netbuf, nbidx, timer);
		nbidx = 0;
	}
	return r;
}
			


#ifdef		EXECUTABLE

int main(int argc, char **argv)
{
	int 	sock, num;
	char	buf[80];
	
	if (argc < 2)  {
		printf("Usage: %s portnum\n", *argv);
		return 1;
	}
	num = atoi(argv[1]);
	if (num < 1)  {
		printf("Bad port number\n");
		return 2;
	}
	sock = OpenSocket("localhost", num);
	if (sock < 0)  {
		printf("open socket failed\n");
		return 2;
	}
	while(1) {
		num = socketReadLine(sock, buf, 80, 0);
		if (num < 0)  return num;
		socketWrite(1, buf, num, 0);
	}
	return 0;
}

#endif

